//
//    CalcEm - Calculator Emulator
//    Copyright (C) 1998  David Eriksson - david@2goodsoft.com
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//////////////////////////////////////////////////////////////////////////////////
// Project Name: [ Calculator Emulator (CalcEm) ]
// Source File:  [ CCalculator class implementation ]
// Author:       [ David Eriksson - david@2goodsoft.com ]
// Revision:     [ 0.5 ]
//////////////////////////////////////////////////////////////////////////////////
// Calculator.cpp: implementation of the CCalculator class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "calcem.h"
#include "Calculator.h"
#include "Link.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//
// The one and only calculator object!!!
//
CCalculator * g_pCalculator = NULL;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CCalculator::CCalculator()
{
	m_pRom = m_pMemory = m_pScreen = NULL;
	m_hLinkThread = m_hEmulatorThread = NULL;
	m_LinkSending = m_LinkActive = false;
	m_RefreshOnInt = true;
	m_hInterrupt = ::CreateEvent(NULL, TRUE, FALSE, NULL);
	m_hInterruptNoWait = ::CreateEvent(NULL, TRUE, TRUE, NULL);
	m_Suspend = 0;
}

CCalculator::~CCalculator()
{
	ASSERT( !m_hEmulatorThread );

	if( m_pScreen )
		delete m_pScreen;

	if( m_pMemory )
		delete m_pMemory;

	if( m_pRom )
		delete m_pRom;
}

//----------------------------------------------------------------------------
//
// Get screen as bitmap
//
//----------------------------------------------------------------------------
CBitmap * CCalculator::GetScreenBitmap()
{
	CBitmap * pbmp = new CBitmap;

	VERIFY( pbmp->CreateBitmap(m_cxScreen, m_cyScreen, 1, 1, m_pScreen) );

	return pbmp;
}

//----------------------------------------------------------------------------
//
// Load ROM
//
//----------------------------------------------------------------------------
bool CCalculator::LoadROM(const char * pszFile)
{
	CFile file(pszFile, CFile::modeRead);
	
	if( file.GetLength() != m_RomSize )
		return false;

	m_pRom = new unsigned char[m_RomSize];
	return file.Read(m_pRom, m_RomSize) == m_RomSize;
}

//----------------------------------------------------------------------------
//
// Start calculator
//
//----------------------------------------------------------------------------
bool CCalculator::Start()
{
	if( m_hEmulatorThread )
		return false;
	
	// Set ROM page
	memcpy(m_pMemory + 0x4000, m_pRom + (m_RomPageNo << 0xE), 0x4000);

	// Initiate Z80 emulator
	InitZ80(m_Regs);

	// Start thread
	m_KeepAlive = true;
	CWinThread * pThread = AfxBeginThread(EmulatorThread, this);

	// Save thread handle
	m_hEmulatorThread = pThread->m_hThread;
	m_EmulatorThreadID = pThread->m_nThreadID;

	return true;
}

//----------------------------------------------------------------------------
//
// Stop calculator
//
//----------------------------------------------------------------------------
bool CCalculator::Stop()
{
	CWaitCursor wait;

	m_KeepAlive = false;

	// Assure thread is running
	while( ResumeThread(m_hEmulatorThread) != 0xffffffff )
		;

	StopLink();

	::WaitForSingleObject(m_hEmulatorThread, INFINITE);
	m_hEmulatorThread = NULL;

	return true;
}

//----------------------------------------------------------------------------
//
// Reset calculator
//
//----------------------------------------------------------------------------
bool CCalculator::Reset()
{
	ResetZ80(&m_Regs);

	ASSERT(m_pMemory);
	ASSERT(m_pRom);

	// Init memory
	memcpy(m_pMemory, m_pRom, 0x4000);				// first 4000h ROM
//	memcpy(m_pMemory + 0x4000, m_pRom, 0x4000);		// next 4000h ROM
	memset(m_pMemory + 0x8000, 0xFF, 0x8000);		// RAM

	m_LinkState = 0;
	m_RomPageNo = m_RomPage = 0;

	return true;
}

//----------------------------------------------------------------------------
//
// Suspend emulation thread
//
//----------------------------------------------------------------------------
DWORD CCalculator::Suspend()
{
	TRACE("Suspending (%i)...", m_Suspend);
	if( m_Suspend++ == 0 )
	{
#ifdef x_DEBUG
		if( m_LinkActive )
			__asm int 3;
#endif

		// Stop thread execution at next instruction
		m_KeepAlive = false;

		// Wait for thread to die
		::WaitForSingleObject(m_hEmulatorThread, INFINITE);
		m_hEmulatorThread = NULL;

		// Stop execution at next interrupt...
//		::ResetEvent(m_hInterruptNoWait);
//		::WaitForSingleObject(m_hInterrupt, INFINITE);

//		::SuspendThread(m_hEmulatorThread);
	}
	TRACE("Done!\n");

	return m_Suspend;
}

//----------------------------------------------------------------------------
//
// Resume emulation thread
//
//----------------------------------------------------------------------------
DWORD CCalculator::Resume()
{
	if( !m_Suspend )
	{
		TRACE("Attempt to resume when running!\n");
//		ASSERT(m_Suspend);
		return (DWORD)-1;	
	}
	TRACE("Resume (%i)\n", m_Suspend);
	if( --m_Suspend == 0 )
	{
		// Clear keyboard buffer
		ZeroMemory(m_VKeys, sizeof(m_VKeys));

		// Start thread again
		m_KeepAlive = true;
		CWinThread * pThread = AfxBeginThread(EmulatorThread, this);

		// Save thread handle
		m_hEmulatorThread = pThread->m_hThread;
		m_EmulatorThreadID = pThread->m_nThreadID;

		SetInputThread( AfxGetApp()->m_nThreadID, TRUE );

		// Continue...
//		::SetEvent(m_hInterruptNoWait);
//		::ResetEvent(m_hInterrupt);
//		::ResumeThread(m_hEmulatorThread);
	}

	return m_Suspend;
}

//----------------------------------------------------------------------------
//
// Z80 Interrupt
//
//----------------------------------------------------------------------------
unsigned short CCalculator::Interrupt()
{
//	::SetEvent(m_hInterrupt);
//	::WaitForSingleObject(m_hInterruptNoWait, INFINITE);

	// Get keyboard state
	VERIFY( GetKeyboardState(m_VKeys) );

	// Update screen
	if( m_RefreshOnInt )
		::PostMessage(m_hwndScreen, wm_Refresh, 0, 0);

	//
	// Any return value except 0xffff will do as we are in interrupt mode 1
	//
	return 0x0000;
}

//----------------------------------------------------------------------------
//
// Get Z80 registers
//
//----------------------------------------------------------------------------
reg * CCalculator::GetRegs()
{
	Z80_GetRegs(&m_Regs);

	return &m_Regs;
}

//----------------------------------------------------------------------------
//
// Set Z80 registers
//
//----------------------------------------------------------------------------
void CCalculator::SetRegs(reg Regs)
{
	m_Regs = Regs;
	Z80_SetRegs(&m_Regs);
}

/*

    RECEIVE
	-------
	5. Input() = line
	4. Output(~line)
	3. Input() = ~line
	2. Output(Red | White)
	1. Input() = Red | White

	SEND
	----
	4. Output(line)
	3. Input() = 0
	2. Output(Red | White)
	1. Input() = Red | White

*/


//----------------------------------------------------------------------------
//
// Link port output
//
//----------------------------------------------------------------------------
void CCalculator::LinkOutput(unsigned char Data)
{
	//
	// Link output:
	// 0 - both
	// 1 - white
	// 2 - red
	//
	m_LinkOutput = Data & 3;
	
	//
	// Link state:
	// 1 - red
	// 2 - white
	// 3 - both
	//
	m_LinkState = m_LinkOutput ^ 3;

	if( m_LinkActive )
	{
#ifdef x_DEBUG
		reg Regs;
		Z80_GetRegs(&Regs);

		TRACE_PC_83(Regs.PC.W);
		TRACE("Link output: %02X\n", m_LinkOutput);
#endif

		//
		// Prevent ROM calls from disturbing the link port activity!!!
		// Dirty, but required!
		//
		if( m_Model == 0x83 )
		{
			reg Regs;
			Z80_GetRegs(&Regs);

			if( Regs.PC.W < 0x4000 )
				return;
		}

		m_LinkPort.LinkOutput();
	}
}

//----------------------------------------------------------------------------
//
// Link port input
//
//----------------------------------------------------------------------------
unsigned char CCalculator::LinkInput()
{
	if( m_LinkActive )
	{
#ifdef x_DEBUG
		reg Regs;
		Z80_GetRegs(&Regs);

		TRACE_PC_83(Regs.PC.W);
		TRACE("Link input\n");
#endif

		//
		// Prevent ROM calls from disturbing the link port activity!!!
		// Dirty, but required!
		//
		if( m_Model == 0x83 )
		{
			reg Regs;
			Z80_GetRegs(&Regs);

			if( Regs.PC.W < 0x4000 )
				return m_LinkState;
		}

		m_LinkPort.LinkInput();
	}

	return m_LinkState;
}

bool CCalculator::SendFiles(CStringArray * pFiles, HWND hProgressWnd)
{
	if( m_LinkActive )
	{
		return false;
//		if( ::WaitForSingleObject(m_hLinkThread, 0) != WAIT_OBJECT_0  )
//		if( ::WaitForSingleObject(m_hLinkThread, INFINITE) != WAIT_OBJECT_0  )
//			return false;
	}

//	char * pszFileName = new char [strlen(pszFile)+1];
//	strcpy(pszFileName, pszFile);
//	CWinThread * pThread = AfxBeginThread(SendToEmulatorThread, (PVOID)pszFileName);

	LINK_THREAD_PARAMS * pltp = new LINK_THREAD_PARAMS;
	pltp->pszFile = NULL;
	pltp->hProgressWnd = hProgressWnd;
	pltp->pFiles = pFiles;

	CWinThread * pThread = AfxBeginThread(SendToEmulatorThread, (PVOID)pltp);

	m_hLinkThread = pThread->m_hThread;
	TRACE("Link thread: 0x%08X\n", pThread->m_nThreadID);

	return true;
}

void CCalculator::StopLink()
{
	if( m_LinkActive )
	{
		CWaitCursor wait;

		TRACE("Link is active, nuking...\n");
		m_LinkPort.m_Dir = Cancel;

		::WaitForSingleObject(m_hLinkThread, INFINITE);
		m_LinkActive = false;
	}
}

BOOL CCalculator::SetInputThread(DWORD idThread, BOOL fAttach)
{
#if 1
	// Vital for keyboard input!
	return ::AttachThreadInput(
		m_EmulatorThreadID,
		idThread, // ::GetCurrentThreadId(),
		fAttach );
#else
	return FALSE;
#endif
}
