//
//    CalcEm - Calculator Emulator
//    Copyright (C) 1998  David Eriksson - david@2goodsoft.com
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//////////////////////////////////////////////////////////////////////////////////
// Project Name: [ Calculator Emulator (CalcEm) ]
// Source File:  [ CTi82 class implementation ]
// Author:       [ David Eriksson - david@2goodsoft.com ]
// Revision:     [ 0.5 ]
//////////////////////////////////////////////////////////////////////////////////
// Ti82.cpp: implementation of the CTi82 class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "calcem.h"
#include "Ti82.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CTi82::CTi82()
{
	m_Model = 0x82;
	m_RomSize = 0x20000;
}

CTi82::~CTi82()
{
	// The Ti82 object has to be alive during Stop()...
	if( m_hEmulatorThread )
		Stop();
}

//----------------------------------------------------------------------------
//
// Reset calculator
//
//----------------------------------------------------------------------------
bool CTi82::Reset()
{
	if( !CMiniCalculator::Reset() )
		return false;

	// Dirty!
	if( *(unsigned long *)(m_pRom + 0x646) == 0x04D3163E )
	{
		m_Regs.PC.W = 0x646;
	}
	else
	{
		return false;
	}

	// Set Interrupt Mode 1 - required!
	m_Regs.IFF = 2;

	return true;
}

//----------------------------------------------------------------------------
//
// Z80 Interrupt
//
//----------------------------------------------------------------------------
unsigned short CTi82::Interrupt()
{
	CMiniCalculator::Interrupt();

	// Some pc keys are mapped to the same calculator key
	COMBINE2( 'M', VK_DIVIDE );
	COMBINE3( 'O', VK_NUMPAD7, '7' );
	COMBINE3( 'P', VK_NUMPAD8, '8' );
	COMBINE3( 'Q', VK_NUMPAD9, '9' );
	COMBINE2( 'R', VK_MULTIPLY );
	COMBINE3( 'T', VK_NUMPAD4, '4' );
	COMBINE3( 'U', VK_NUMPAD5, '5' );
	COMBINE3( 'V', VK_NUMPAD6, '6' );
	COMBINE2( 'W', VK_SUBTRACT );
	COMBINE3( 'Y', VK_NUMPAD1, '1' );
	COMBINE3( 'Z', VK_NUMPAD2, '2' );
	COMBINE2( '3', VK_NUMPAD3 );
	COMBINE3( ' ', VK_NUMPAD0, '0' );
	COMBINE2( VK_BACK, VK_LEFT );
	COMBINE2( VK_INSERT, VK_DELETE );

	// Tab key is ON
	m_ONKeyDown = m_VKeys[VK_TAB] >> 7;

	//
	// Any return value except 0xffff will do as we are in interrupt mode 1
	//
	return 0x0000;
}

#ifdef _DEBUG
	static int odbg = 0;
	static int idbg = 0;
#endif

//----------------------------------------------------------------------------
//
// I/O port input
//
//----------------------------------------------------------------------------
unsigned char CTi82::Input(unsigned char Port)
{
#ifdef _DEBUG
	reg Regs;
	Z80_GetRegs(&Regs);

	if( idbg && Port != PORT_ROMPAGE)
	{
		TRACE_PC(Regs.PC.W);
		TRACE("Input %X\n", Port);
	}
#endif

	switch(Port)
	{
	case PORT_LINK:
		// Prepare input
		LinkInput();

		//
		// Bits:
		// 0-1 - State
		// 2-3 - Output state (bit 2-3)
		// 4-7 - Always 0?
		//
		return m_LinkOutput << 2 | m_LinkState;

	case PORT_KEYBOARD:
		return KeyboardInput();

	case PORT_ROMPAGE:
		return m_RomPage;

	case PORT_STATUS:
		//
		// Bits:
		// 0 - Set if ON key is down
		// 1 - On interrupt and ON key not down, this bit set causes:
		//     Busy counter update, Keyboard input, Cursor update, APD update
		// 2 - Function unknown, never set
		// 3 - Set if ON key is up (according to 82-ports.txt)
		//
		if( m_ONKeyDown )
			return (m_MaskON | 2);
		else
			return (8 | 2);				

	case PORT_INTCTRL:
#ifdef _DEBUG
		TRACE_PC(Regs.PC.W);
		TRACE("Input from \"turbo\" port.\n");
#endif
		return 0xFF;

	case PORT_DISPDATA:
		return DisplayInput();

	case PORT_DISPCTRL:
		// undefined

	default:
		ASSERT(0);	
		return 0xFF;
	}
}

//----------------------------------------------------------------------------
//
// I/O port output
//
//----------------------------------------------------------------------------
void CTi82::Output(unsigned char Port, unsigned char Data)
{
#ifdef _DEBUG
	reg Regs;
	Z80_GetRegs(&Regs);

	if( odbg && Port != PORT_ROMPAGE )
	{
		TRACE_PC(Regs.PC.W);
		TRACE("Output %X, %02X\n", Port, Data);
	}
#endif

	switch(Port)
	{
	case PORT_LINK:
		//
		// Bits:
		// 0-1 - Always 00?
		// 2-5 - State (abab)
		// 6-7 - Always 11?
		//
		LinkOutput(Data >> 2);
		break;

	case PORT_KEYBOARD:
		m_KeyboardMask = Data;
		break;

	case PORT_ROMPAGE:
		m_RomPage = Data;

		m_RomPageNo = m_RomPage & 7;
		memcpy(m_pMemory + 0x4000, m_pRom + (m_RomPageNo << 0xE), 0x4000);
		break;

	case PORT_STATUS:
		//
		// Bits:
		// 0 - Mask ON key down
		// 1 - "VBI" (Vertical Blanking Interrupt?) according to 82-ports.txt
		// 2 - Unused? (82-ports.txt has a theory)
		// 3 - Probably display power: 0 - off, 1 - on (not like 82-ports.txt)
		//
		// On each interrupt this is written to the status port:
		// (a) 1010b - if ON key is down
		//     1001b - if ON key is up
		// (b) 1011b - always
		//
		m_MaskON = Data & 1;
#ifdef _DEBUG
		if( Data & 4 )
			TRACE("%04X  Status = %02X - Unexpected bit set!\n", Regs.PC.W, Data);
#endif
		break;

	case PORT_INTCTRL:
//		TRACE("%04X  Interrupt control = %02X\n", Regs.PC.W, Data);
		// ignore
		break;

	case PORT_DISPCTRL:
		DisplayControl(Data);
		break;

	case PORT_DISPDATA:
		DisplayOutput(Data);
		break;

	default:
		//
		// No other ports should ever be accessed
		//
		ASSERT(0);	
	}
}


/*

	Calculator to computer key map
	==============================

	1 to 9		1 to 9, Numeric keypad 1 to 9
	A to Z		Keyboard A to Z

	-			W, Numeric keypad -
	(			K
	(-)			F10
	)			L
	*			R, Numeric keypad *
	,			J
	.			Numeric keypad decimal sign
	/			M, Numeric keypad /
	^			H
	+			Numeric keypad +
	2nd			Ctrl
	Alpha		Shift
	Clear		Escape
	Cos			F
	Del			Delete, Insert
	Enter		Return, Enter
	Graph		F5
	Ln			S
	Log			N
	Math		A
	Matrx		B
	Mode		F6
	On			Tab
	Prgm		C
	Sin			E
	Stat		F8
	Sto>		X
	Tan			G
	Trace		F4
	Vars		F9
	Window		F2
	X,T,theta	F7
	X^-1		D
	X^2			I
	Y=			F1
	Zoom		F3

*/

//
//	Index in this table + 1 = keyboard scan code...
//
static byte KeyMap[7][8] = 
{
	// down, left, right, up
	{ VK_DOWN, VK_LEFT, VK_RIGHT, VK_UP, 0, 0, 0, 0 },
	
	// enter, add, subtract, multiply, divide, power, clear
	{ VK_RETURN, VK_ADD, VK_SUBTRACT, VK_MULTIPLY, VK_DIVIDE, 'H', VK_ESCAPE, 0 },

	// (-), 3, 6, 9, ), Tan, Vars
	{ VK_F10, VK_NUMPAD3, VK_NUMPAD6, VK_NUMPAD9, 'L', 'G', VK_F9, 0 },

	// decimal, 2, 5, 8, (, Cos, Prgm, Stat, 
	{ VK_DECIMAL, VK_NUMPAD2, VK_NUMPAD5, VK_NUMPAD8, 'K', 'F', 'C', VK_F8 },
	
	// 0, 1, 4, 7, comma, sin, matrx, x,t,theta
	{ VK_NUMPAD0, VK_NUMPAD1, VK_NUMPAD4, VK_NUMPAD7, 'J', 'E', 'B', VK_F7 },

	// sto>, ln, log, x^2, x^-1, math, alpha
	{ 0, 'X', 'S', 'N', 'I', 'D', 'A', VK_SHIFT },

	// graph, trace, zoom, window, y=, 2nd, mode, del
	{ VK_F5, VK_F4, VK_F3, VK_F2, VK_F1, VK_CONTROL, VK_F6, VK_DELETE },
};


//----------------------------------------------------------------------------
//
// Keyboard Input
//
//----------------------------------------------------------------------------
unsigned char CTi82::KeyboardInput()
{
	// Disable keyboard while ALT is pressed!
	if( m_VKeys[VK_MENU] & 128 )
		return 0xff;

	//
	// It seems like the mask has to be inverted to work nicely
	//
	BYTE Mask = ~m_KeyboardMask;
	BYTE Key;

	BYTE KeyboardPress = 0;

	// Loop through the low 7 bits in mask
	for( BYTE i = 0; i < 7; i++ )
	{
		if( Mask & 1 )
		{
			// See if any key in this mask is pressed
			for( BYTE j = 0; j < 8; j++ )
			{
				if( Key = KeyMap[i][j] )
					if( m_VKeys[Key] & 128 )
						KeyboardPress |= (1 << j);
			}
		}

		// Next bit...
		Mask >>= 1;
	}

	return ~KeyboardPress;
}
 
bool CTi82::LoadRAM(const char * pszFile)
{
	CWaitCursor cur;
	CFile file(pszFile, CFile::modeRead);

	unsigned Version = 0, cFile = file.GetLength();

	switch( cFile )
	{
	case 0x8327:
		// TI-82 saved image from TI-8X emulator v0.1
		Version = 0x0100;
		break;

	case 0x832b:
		// TI-82 saved image from TI-8X emulator v0.2b1
		Version = 0x0201;
		break;

	default:
		return false;
	}

	// RAM image
	file.Read(m_pMemory + 0x8000, 0x8000);

	// State
	unsigned cState = file.GetLength() & 0xFF;
	unsigned char * pState = new unsigned char[cState];

	file.Read(pState, cState);
	
	// Screen image
	file.Read(m_pScreen, 768);

	//
	// Process state
	//

	// These register pairs are stored "backwards"
	// AF, BC, DE, HL, IX, IY, PC, SP, AF', BC', DE', HL'
	_swab((char *)pState, (char *)&m_Regs, 0x18);
	//
	// You don't know what this function does? Read the fine manual:
	//
	// void _swab( char *src, char *dest, int n );
	//
	// The _swab function copies n bytes from src, swaps each pair of
	// adjacent bytes, and stores the result at dest. The integer n
	// should be an even number to allow for swapping. _swab is typically 
	// used to prepare binary data for transfer to a machine that uses a 
	// different byte order.
	//

	// IFF1
	if( pState[0x18] )
		m_Regs.IFF |= 1;

	// IFF2
	if( pState[0x19] )
		m_Regs.IFF |= 64;

	// HALT
	if( pState[0x1A] )
		m_Regs.IFF |= 128;

	// Interrupt mode (0 to 2)
	if( !pState[0x1B] )
	{
		TRACE("Warning! IM 0 selected, changing to IM 1!\n");
		pState[0x1B] = 1;
	}
	m_Regs.IFF = (m_Regs.IFF & ~6) | (pState[0x1B] << 1);

	// I
	m_Regs.I = pState[0x1C];

	// R, R2: two bytes

	// Keyboard mask
	m_KeyboardMask = pState[0x1F];

	// ON mask
	m_MaskON = pState[0x20];

	if( Version == 0x0100 )
	{
		// Unknown/unused

		// ROM page
		m_RomPage = pState[0x22];
		
		// Display stuff
		m_DispX = pState[0x23];
		m_DispY = pState[0x24];
		m_DisplayMove = pState[0x25];
		m_DisplayMode = pState[0x26];
	}
	else if( Version == 0x0201 )
	{
		// "LinkReg" (0x21)

		// "LCDmask" (0x22)

		// "LCDon" (0x23)

		// Contrast (0x24)

		// ROM page
		m_RomPage = pState[0x25];

		// Display stuff
		m_DispX = pState[0x26];
		m_DispY = pState[0x27];
		m_DisplayMove = pState[0x28];
		m_DisplayMode = pState[0x29];

		// Hardware scroll (0x2a)
	}

	m_RomPageNo = m_RomPage & 7;

	ICount = 0;		// Immediate interrupt!

	delete pState;
	return true;
}

bool CTi82::StoreRAM(const char * pszFile, unsigned nVersion)
{
	CWaitCursor cur;

	Suspend();
	GetRegs();

	unsigned cState;
	
	if( nVersion == 0x0100 )
		cState = 0x27;
	else if( nVersion == 0x0201 )
		cState = 0x2b;
	else
		return false;

	unsigned char * pState = new unsigned char[cState];

	// These register pairs are stored "backwards"
	// AF, BC, DE, HL, IX, IY, PC, SP, AF', BC', DE', HL'
	_swab((char *)&m_Regs, (char *)pState, 24);
	//
	// You don't know what this function does? Read the fine manual:
	//
	// void _swab( char *src, char *dest, int n );
	//
	// The _swab function copies n bytes from src, swaps each pair of
	// adjacent bytes, and stores the result at dest. The integer n
	// should be an even number to allow for swapping. _swab is typically 
	// used to prepare binary data for transfer to a machine that uses a 
	// different byte order.
	//

	// IFF1
	pState[0x18] = !!(m_Regs.IFF & 1);

	// IFF2
	pState[0x19] = !!(m_Regs.IFF & 64);

	// HALT
	pState[0x1A] = !!(m_Regs.IFF & 128);

	// Interrupt mode (0 to 2)
	pState[0x1B] = (m_Regs.IFF >> 1) & 3;

	// I
	pState[0x1C] = m_Regs.I;

	// R, R2: two bytes
	pState[0x1D] = 0;
	pState[0x1E] = 0;

	// Keyboard mask
	pState[0x1F] = m_KeyboardMask;

	// ON mask
	pState[0x20] = m_MaskON;

	if( nVersion == 0x0100 )
	{
		// Unknown/unused
		pState[0x21] = 0;

		// ROM page
		pState[0x22] = m_RomPage;

		// Display stuff
		pState[0x23] = m_DispX;
		pState[0x24] = m_DispY;
		pState[0x25] = m_DisplayMove;
		pState[0x26] = m_DisplayMode;
	}
	else if( nVersion == 0x0201 )
	{
		// "LinkReg"
		pState[0x21] = 0;		// 00

		// "LCDmask" (0x22)
		pState[0x22] = 0;		// 02

		// Display power
		pState[0x23] = 0;		// 08

		// Contrast
		pState[0x24] = 0;		// 0D

		// ROM page
		pState[0x25] = m_RomPage;

		// Display stuff
		pState[0x26] = m_DispX;
		pState[0x27] = m_DispY;
		pState[0x28] = m_DisplayMove;
		pState[0x29] = m_DisplayMode;
		pState[0x2a] = 0;					// Hardware scroll
	}
		
	CFile file(pszFile, CFile::modeCreate | CFile::modeWrite );

	// RAM image
	file.Write(m_pMemory + 0x8000, 0x8000);

	// State
	file.Write(pState, cState);

	// Screen image
	file.Write(m_pScreen, 0x300);

	// Continue execution
	Resume();

	delete pState;
	return true;

}
