//
//    CalcEm - Calculator Emulator
//    Copyright (C) 1998  David Eriksson - david@2goodsoft.com
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//////////////////////////////////////////////////////////////////////////////////
// Project Name: [ Calculator Emulator (CalcEm) ]
// Source File:  [ CTi83 class implementation ]
// Author:       [ David Eriksson - david@2goodsoft.com ]
// Revision:     [ 0.5 ]
//////////////////////////////////////////////////////////////////////////////////
// Ti83.cpp: implementation of the CTi83 class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "calcem.h"
#include "Ti83.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CTi83::CTi83()
{
	m_Model = 0x83;
	m_RomSize = 0x40000;

	m_RomPageNo = 0;
}

CTi83::~CTi83()
{
	// The Ti83 object has to be alive during Stop()...
	if( m_hEmulatorThread )
		Stop();
}

bool CTi83::Reset()
{
	if( !CMiniCalculator::Reset() )
		return false;

	// Dirty!
	if( *(unsigned long *)(m_pRom + 0x6ce) == 0x04D3163E )
	{
		m_Regs.PC.W = 0x6ce;
	}
	else if( *(unsigned long *)(m_pRom + 0x6f6) == 0x04D3163E )
	{
		m_Regs.PC.W = 0x6f6;
	}
	else
	{
		return false;
	}

#ifdef DEBUG
	Trap = m_Regs.PC.W;
#endif

	// Set Interrupt Mode 1 - required!
	m_Regs.IFF = 2;

	return true;
}

unsigned char CTi83::Input(unsigned char Port)
{
#ifdef _DEBUG
	reg Regs;
	Z80_GetRegs(&Regs);

//	if( Port != 0 && Port != 2 )
//	{
//		TRACE_PC(Regs.PC.W);
//		TRACE("Input(%02X)\n", Port);
//	}
#endif

	switch(Port)
	{
	case PORT_LINK:
		LinkInput();
		
		//
		// Bits:
		// 0-1 - Output state?
		// 2-3 - Input state
		// 4   - ROM Page bit
		// 5-7 - Always 0?
		//
		return (m_RomPageNo & 8) << 1 | m_LinkState << 2 | m_LinkOutput;

	case PORT_KEYBOARD:
		return KeyboardInput();

	case PORT_ROMPAGE:
		return m_RomPage;

	case PORT_STATUS:
#ifdef x_DEBUG
		if( m_LinkActive )
			TRACE("%04X  Status port input while link is active!\n", Regs.PC.W);
#endif
		//
		// Bits:
		// 0   - Set if ON key is down and ON key is trapped
		// 1   - Update things (keyboard etc)
		// 2   - Unknown, but used
		// 3   - Set if ON key is up
		// 4-7 - Unknown
		return (m_ONKeyDown ? m_MaskON : 8) |
			//2; //((m_Status & 4) ? 2 : 0);
			(m_LinkSending ? 0 : 2);

	case PORT_INTCTRL:
		return 0xFF;

	case PORT_DISPDATA:
		return DisplayInput();

	case PORT_DISPCTRL:
		// undefined

	default:
		ASSERT(0);	
		return 0xFF;
	}
}

void CTi83::Output(unsigned char Port, unsigned char Data)
{
#ifdef _DEBUG
	reg Regs;
	Z80_GetRegs(&Regs);

//	TRACE_PC_83(Regs.PC.W);
//	TRACE("Output(%02X, %02X)\n", Port, Data);
#endif

	unsigned char temp;

	switch(Port)
	{
	case PORT_LINK:
		//
		// Bits:
		// 0-1 - Output state
		// 2-3 - Always 0?
		// 4   - ROM Page bit
		// 6-7 - Always 11
		//

		LinkOutput(Data);

		temp = (m_RomPageNo & 7) | (Data & 16) >> 1;
		if( m_RomPageNo != temp )
		{
			m_RomPageNo = temp;
			memcpy(m_pMemory + 0x4000, m_pRom + (m_RomPageNo << 0xE), 0x4000);
		}
		break;

	case PORT_KEYBOARD:
		m_KeyboardMask = Data;
		break;

	case PORT_ROMPAGE:
		m_RomPage = Data;

		temp = (m_RomPageNo & 8) | (Data & 7);
		if( m_RomPageNo != temp )
		{
			m_RomPageNo = temp;
			memcpy(m_pMemory + 0x4000, m_pRom + (m_RomPageNo << 0xE), 0x4000);
		}
		break;

	case PORT_STATUS:
//		m_Status = Data;
		m_MaskON = Data & 1;
		break;

	case PORT_INTCTRL:
		// ignore
#ifdef _DEBUG
		if( Data != 0x16 )
			TRACE("%04X  Interrupt control = %02X\n", Regs.PC.W, Data);
#endif
		break;

	case PORT_DISPCTRL:
		DisplayControl(Data);
		break;

	case PORT_DISPDATA:
		DisplayOutput(Data);
		break;

	default:
		//
		// No other ports should ever be accessed
		//
		ASSERT(0);	
	}
}

unsigned short CTi83::Interrupt()
{
	CMiniCalculator::Interrupt();

	// Some pc keys are mapped to the same calculator key
	COMBINE2( 'M', VK_DIVIDE );
	COMBINE3( 'O', VK_NUMPAD7, '7' );
	COMBINE3( 'P', VK_NUMPAD8, '8' );
	COMBINE3( 'Q', VK_NUMPAD9, '9' );
	COMBINE2( 'R', VK_MULTIPLY );
	COMBINE3( 'T', VK_NUMPAD4, '4' );
	COMBINE3( 'U', VK_NUMPAD5, '5' );
	COMBINE3( 'V', VK_NUMPAD6, '6' );
	COMBINE2( 'W', VK_SUBTRACT );
	COMBINE3( 'Y', VK_NUMPAD1, '1' );
	COMBINE3( 'Z', VK_NUMPAD2, '2' );
	COMBINE2( '3', VK_NUMPAD3 );
	COMBINE3( ' ', VK_NUMPAD0, '0' );
	COMBINE2( VK_BACK, VK_LEFT );
	COMBINE2( VK_INSERT, VK_DELETE );

	// Tab key is ON
	m_ONKeyDown = m_VKeys[VK_TAB] >> 7;

	//
	// Any return value except 0xffff will do as we are in interrupt mode 1
	//
	return 0x0000;
}

//
//	Index in this table + 1 = keyboard scan code...
//
static byte KeyMap[7][8] = 
{
	// down, left, right, up
	{ VK_DOWN, VK_LEFT, VK_RIGHT, VK_UP, 0, 0, 0, 0 },
	
	// enter, add, subtract, multiply, divide, power, clear
	{ VK_RETURN, VK_ADD, VK_SUBTRACT, VK_MULTIPLY, VK_DIVIDE, 'H', VK_ESCAPE, 0 },

	// (-), 3, 6, 9, ), Tan, Vars
	{ VK_F10, VK_NUMPAD3, VK_NUMPAD6, VK_NUMPAD9, 'L', 'G', VK_F9, 0 },

	// decimal, 2, 5, 8, (, Cos, Prgm, Stat, 
	{ VK_DECIMAL, VK_NUMPAD2, VK_NUMPAD5, VK_NUMPAD8, 'K', 'F', 'C', VK_F8 },
	
	// 0, 1, 4, 7, comma, sin, matrx, x,t,theta
	{ VK_NUMPAD0, VK_NUMPAD1, VK_NUMPAD4, VK_NUMPAD7, 'J', 'E', 'B', VK_F7 },

	// sto>, ln, log, x^2, x^-1, math, alpha
	{ 0, 'X', 'S', 'N', 'I', 'D', 'A', VK_SHIFT },

	// graph, trace, zoom, window, y=, 2nd, mode, del
	{ VK_F5, VK_F4, VK_F3, VK_F2, VK_F1, VK_CONTROL, VK_F6, VK_DELETE },
};

//----------------------------------------------------------------------------
//
// Keyboard Input
//
//----------------------------------------------------------------------------
unsigned char CTi83::KeyboardInput()
{
	// Disable keyboard while ALT is pressed!
	if( m_VKeys[VK_MENU] & 128 )
		return 0xff;

	//
	// It seems like the mask has to be inverted to work nicely
	//
	BYTE Mask = ~m_KeyboardMask;
	BYTE Key;

	BYTE KeyboardPress = 0xff;

	// Loop through the low 7 bits in mask
	for( BYTE i = 0; i < 7; i++ )
	{
		if( Mask & 1 )
		{
			// See if any key in this mask is pressed
			for( BYTE j = 0; j < 8; j++ )
			{
				if( Key = KeyMap[i][j] )
					if( m_VKeys[Key] & 128 )
						KeyboardPress ^= (1 << j);
			}
		}

		// Next bit...
		Mask >>= 1;
	}

	return KeyboardPress;
}
 
bool CTi83::LoadRAM(const char * pszFile)
{
	CWaitCursor cur;
	CFile file(pszFile, CFile::modeRead);

	unsigned Version = 0, cFile = file.GetLength();

	switch( cFile )
	{
	case 0x8327:
		// TI-83 saved image from TI-8X emulator v0.1
		Version = 0x00100;
		break;

	case 0x832c:
		// TI-83 saved image from TI-8X emulator v0.2b1
		Version = 0x00201;
		break;

	case 0x8330:
		// EMU83 : TI-83 Emulator Version 0.2
		Version = 0x10200;
		break;

	default:
		return false;
	}

	// RAM image
	file.Read(m_pMemory + 0x8000, 0x8000);

	// State
	unsigned cState = file.GetLength() & 0xFF;
	unsigned char * pState = new unsigned char[cState];

	switch( Version >> 16 )
	{
	case 0:
		//
		// TI8XEMU...
		//
		file.Read(pState, cState);
		
		// Screen image
		file.Read(m_pScreen, 768);

		//
		// Process state
		//

		// These register pairs are stored "backwards"
		// AF, BC, DE, HL, IX, IY, PC, SP, AF', BC', DE', HL'
		_swab((char *)pState, (char *)&m_Regs, 0x18);
		//
		// You don't know what this function does? Read the fine manual:
		//
		// void _swab( char *src, char *dest, int n );
		//
		// The _swab function copies n bytes from src, swaps each pair of
		// adjacent bytes, and stores the result at dest. The integer n
		// should be an even number to allow for swapping. _swab is typically 
		// used to prepare binary data for transfer to a machine that uses a 
		// different byte order.
		//

		// IFF1
		if( pState[0x18] )
			m_Regs.IFF |= 1;

		// IFF2
		if( pState[0x19] )
			m_Regs.IFF |= 64;

		// HALT
		if( pState[0x1A] )
			m_Regs.IFF |= 128;

		// Interrupt mode (0 to 2)
		//m_Regs.IFF = 1;
		m_Regs.IFF |= (pState[0x1B] << 1);

		// I
		m_Regs.I = pState[0x1C];

		// R, R2: two bytes

		// Keyboard mask
		Output(PORT_KEYBOARD, pState[0x1f]);

		// ON mask
		m_MaskON = pState[0x20];

		if( Version == 0x0100 )
		{
			// Unknown/unused

			// ROM page
			m_RomPage = pState[0x22];
			
			// Display stuff
			m_DispX = pState[0x23];
			m_DispY = pState[0x24];
			m_DisplayMove = pState[0x25];
			m_DisplayMode = pState[0x26];
		}
		else if( Version == 0x0201 )
		{
			// "LinkReg" (0x21)

			// "LCDmask" (0x22)

			// "LCDon" (0x23)

			// Contrast (0x24)

			// Link port
			Output(PORT_LINK, pState[0x25]);

			// ROM page
			Output(PORT_ROMPAGE, pState[0x26]);

			// Display stuff
			m_DispX = pState[0x27];
			m_DispY = pState[0x28];
			m_DisplayMove = pState[0x29];
			m_DisplayMode = pState[0x2a];

			// Hardware scroll (0x2b)
		}
		break;

	case 1:	{
		// Regs+misc
		file.Read(pState, 0x29);
		
		// Screen image
		file.Read(m_pScreen, 768);

		// Regs+misc
		file.Read(pState + 0x28, 7 );
		
		unsigned short * pStateW = (unsigned short *)pState;

		m_Regs.AF.W = pStateW[0x00];
		m_Regs.BC.W = pStateW[0x01];
		m_Regs.DE.W = pStateW[0x02];
		m_Regs.HL.W = pStateW[0x03];
		m_Regs.PC.W = pStateW[0x04];
		m_Regs.AF1.W = pStateW[0x05];
		m_Regs.BC1.W = pStateW[0x06];
		m_Regs.DE1.W = pStateW[0x07];
		m_Regs.HL1.W = pStateW[0x08];

		// RFlo/RFhi - two words
		
		m_Regs.I = pState[0x16];
		m_Regs.IFF |= pState[0x18] << 1;

		m_Regs.IX.W = pStateW[0x0d];
		m_Regs.IY.W = pStateW[0x0e];

		// IFF1, IFF2
		m_Regs.IFF |= pState[0x1e];
		m_Regs.IFF |= pState[0x1f] << 6;

		m_Regs.SP.W = pStateW[0x10];

		// Current ROM page
		m_RomPageNo = pState[0x22];
		memcpy(m_pMemory + 0x4000, m_pRom + (m_RomPageNo << 0xE), 0x4000);

		// Keypad Mask
		Output(PORT_KEYBOARD, pState[0x23]);
		
		// Interrupt Mask (0x24)
		// Interrupt Status (0x25)
		// Power Register (0x26)
		
		// Link Register
		Output(PORT_LINK, pState[0x27]);

		// Display Contrast (0x28)

		// LCD curX
		m_DispX = pState[0x29];

		// LCD curY
		m_DispY = pState[0x2a];
		
		// LCD depl mode
		m_DisplayMove = pState[0x2b] + 4;

		// LCD HardScroll offset (0x2c)
		// LCD IsOff (0x2d)
		// LCD Last function (0x2e)
		// LCD word Align value (0x2f)
		}
		break;
	}

	delete pState;
	return true;
}

bool CTi83::StoreRAM(const char * pszFile, unsigned nVersion)
{
	CWaitCursor cur;

	Suspend();
	GetRegs();

	unsigned cState;
	
	if( nVersion == 0x0100 )
		cState = 0x27;
	else if( nVersion == 0x0201 )
		cState = 0x2c;
	else
		return false;

	unsigned char * pState = new unsigned char[cState];

	// These register pairs are stored "backwards"
	// AF, BC, DE, HL, IX, IY, PC, SP, AF', BC', DE', HL'
	_swab((char *)&m_Regs, (char *)pState, 24);
	//
	// You don't know what this function does? Read the fine manual:
	//
	// void _swab( char *src, char *dest, int n );
	//
	// The _swab function copies n bytes from src, swaps each pair of
	// adjacent bytes, and stores the result at dest. The integer n
	// should be an even number to allow for swapping. _swab is typically 
	// used to prepare binary data for transfer to a machine that uses a 
	// different byte order.
	//

	// IFF1
	pState[0x18] = !!(m_Regs.IFF & 1);

	// IFF2
	pState[0x19] = !!(m_Regs.IFF & 64);

	// HALT
	pState[0x1A] = !!(m_Regs.IFF & 128);

	// Interrupt mode (0 to 2)
	pState[0x1B] = 0;		// ????
//	pState[0x1B] = (m_Regs.IFF >> 1) & 3;

	// I
	pState[0x1C] = m_Regs.I;

	// R, R2: two bytes
	pState[0x1D] = 0;
	pState[0x1E] = 0;

	// Keyboard mask
	pState[0x1F] = m_KeyboardMask;

	// ON mask
	pState[0x20] = m_MaskON;

	if( nVersion == 0x0100 )
	{
		// Unknown/unused
		pState[0x21] = 0;

		// ROM page
		pState[0x22] = m_RomPage;

		// Display stuff
		pState[0x23] = m_DispX;
		pState[0x24] = m_DispY;
		pState[0x25] = m_DisplayMove;
		pState[0x26] = m_DisplayMode;
	}
	else if( nVersion == 0x0201 )
	{
		// "LinkReg"
		pState[0x21] = 0;

		// "LCDmask" (0x22)
		pState[0x22] = 0;

		// Display power
		pState[0x23] = 0;

		// Contrast
		pState[0x24] = 0;

		// Link port
		pState[0x25] = Input(PORT_LINK);

		// ROM page
		pState[0x26] = Input(PORT_ROMPAGE);

		// Display stuff
		pState[0x27] = m_DispX;
		pState[0x28] = m_DispY;
		pState[0x29] = m_DisplayMove;
		pState[0x2a] = m_DisplayMode;
		pState[0x2b] = 0;					// Hardware scroll
	}
		
	CFile file(pszFile, CFile::modeCreate | CFile::modeWrite );

	// RAM image
	file.Write(m_pMemory + 0x8000, 0x8000);

	// State
	file.Write(pState, cState);

	// Screen image
	file.Write(m_pScreen, 0x300);

	// Continue execution
	Resume();

	delete pState;
	return true;
}
